@@ -23,6 +23,8 @@ import stat
 import time
 import string
 import math
+import threading
+import Queue
 
 import gtk
 import gtk.glade
@@ -33,7 +35,13 @@ import gettext
 
 import puppy
 
-APP_NAME = 'guppy'
+APP_NAME = 'Guppy'
+VERSION = '0.0.2'
+AUTHORS = ['Tony Tsui tsui.tony@gmail.com']
+WEBSITE = 'http://guppy.nongnu.org'
+COPYRIGHT = 'Copyright (C) 2005 Tony Tsui'
+LICENSE = 'GNU Public License'
+
 
 def humanReadableSize(size):
 	div_count = 0
@@ -86,6 +94,13 @@ class FileSystemModel(gtk.ListStore):
 		                             gobject.TYPE_STRING, gobject.TYPE_STRING,
 		                             gobject.TYPE_STRING)
 
+	def find(self, name):
+		for row in self:
+			if name == row[FileSystemModel.NAME_COL]:
+				return row
+				
+		return None
+			
 	def getCWD(self):
 		return self.current_dir
 	
@@ -192,29 +207,49 @@ class PVRFileSystemModel(FileSystemModel):
 
 		# FIXME: Get dir from when Guppy last exited
 		self.current_dir = ''
-		
+
+		self.freespace = 0
+				
 		self.puppy = puppy.Puppy()
 		
 		self.changeDir()
 
 
 	def changeDir(self, dir=None):
-		if len(self) > 0:
-			self.clear()
-			
 		if dir:
-			if dir[0] != '\\':
+			# PVR root dir path is an empty string. Using '\' character does not
+			# work.
+			if dir == '\\':
+				self.current_dir = ''
+			elif len(self.current_dir) > 0 and self.current_dir[-1] != '\\':
 				dir = self.current_dir + '\\' + dir
 		else:
 			dir = self.current_dir
 
 		norm_path = os.path.normpath(dir.replace('\\', '/'))
 		if norm_path != '.':
-			self.current_dir = norm_path.replace('/', '\\')
+			dir = norm_path.replace('/', '\\')
 		else:
-			self.current_dir = '\\'
+			dir = ''
 
-		pvr_files = self.puppy.listDir(self.current_dir)
+		# We can't get a listing during a file transfer
+		try:
+			pvr_files = self.puppy.listDir(dir)
+		except puppy.PuppyBusyError:
+			msg = _('Can not change directory during a file transfer')
+			dialog = gtk.MessageDialog(type=gtk.MESSAGE_INFO,
+			                           buttons=gtk.BUTTONS_CLOSE,
+			                           message_format=msg)
+			response = dialog.run()
+			dialog.destroy()
+
+			return
+
+		self.current_dir = dir
+		
+		# Clear model
+		if len(self) > 0:
+			self.clear()
 
 		for file in pvr_files:
 			# TODO: Set icon based on file type. Use dummy icon for now
@@ -227,8 +262,13 @@ class PVRFileSystemModel(FileSystemModel):
 			self.append(file)
 	
 	def freeSpace(self):
-		total, free = self.puppy.getDiskSpace()
-		return humanReadableSize(free)
+		try:
+			total, free = self.puppy.getDiskSpace()
+			self.freespace = free
+		except:
+			pass
+
+		return humanReadableSize(self.freespace)
 
 class PCFileSystemModel(FileSystemModel):
 	def __init__(self):
@@ -293,6 +333,8 @@ class PCFileSystemModel(FileSystemModel):
 		return humanReadableSize(int(output[3])*1024)
 			
 class GuppyWindow:
+	FREE_SPACE_UPDATE_INTERVAL = 5000 # 5 secs
+	
 	def __init__(self):	
 		# Find out proper way to find glade files
 		guppy_glade_file = 'guppy.glade'
@@ -315,8 +357,6 @@ class GuppyWindow:
 		
 		self.show_hidden = False
 
-		self.transfer_dialog = self.glade_xml.get_widget('transfer_dialog')
-		
 		self.pvr_total_size_label = self.glade_xml.get_widget('pvr_total_size_label')
 		self.pvr_free_space_label = self.glade_xml.get_widget('pvr_free_space_label')
 
@@ -326,8 +366,21 @@ class GuppyWindow:
 		self.pvr_model = PVRFileSystemModel()
 		self.pc_model = PCFileSystemModel()
 		
-		self.free_space_timeout_id = gobject.timeout_add(5000, self.update_free_space)
-		self.update_free_space()
+		self.free_space_timeout_id = gobject.timeout_add(GuppyWindow.FREE_SPACE_UPDATE_INTERVAL, self.update_free_space)
+
+		# Queue to put files to be transferred. The transfer thread gets files
+		# to transfer from this queue.
+		self.transfer_queue = Queue.Queue(0)
+		
+		# Queue to put all files which has been transferred. Files on this queue
+		# are removed when the Transfer Frame Clear button is clicked.
+		self.transfer_complete_queue = Queue.Queue(0)
+		
+		# Create thread to transfer files
+		self.transfer_thread = TransferThread(self)
+		self.transfer_thread.setDaemon(True)
+		self.transfer_thread.start()
+	
 		
 	def initUIManager(self):
 		self.uimanager = gtk.UIManager()
@@ -343,35 +396,34 @@ class GuppyWindow:
 
 		# FIXME: Use a proper icon for Turbo button
 		actiongroup.add_toggle_actions([('Turbo', gtk.STOCK_EXECUTE, 'Tur_bo', None, 'Turbo Transfer', self.on_turbo_toggled),
-		                                ('ShowHidden', None, 'Show Hidden Files', None, 'Show hidden files', self.on_show_hidden_toggled)])
-		                                
+		                                ('ShowHidden', None, 'Show Hidden Files', None, 'Show hidden files', self.on_show_hidden_toggled),
+		                                ('ShowFileTransfer', None, 'Show File Transfer', None, 'Show File Transfer', self.on_show_file_transfer_toggled)])
+
+		# Create reference to ShowFileTransfer action so we can update the
+		# toggle state accordingly when the transfer frame close button is
+		# clicked.
+		self.show_file_transfer_action = actiongroup.get_action('ShowFileTransfer')
 		
 		self.uimanager.insert_action_group(actiongroup, 0)
 		
+		# Create separate action group for upload so sensitivity can be set
+		# according to selection in PC file tree
 		self.upload_actiongrp = gtk.ActionGroup('UploadAction')                                 
 		self.upload_actiongrp.add_actions([('Upload', gtk.STOCK_GO_BACK, '_Upload', None, 'Upload File', self.on_upload_btn_clicked)])
 		self.upload_actiongrp.set_sensitive(False)
 		self.uimanager.insert_action_group(self.upload_actiongrp, 1)
 
+		# Create separate action group for download so sensitivity can be set
+		# according to selection in PVR file tree
 		self.download_actiongrp = gtk.ActionGroup('DownloadAction')                                 
 		self.download_actiongrp.add_actions([('Download', gtk.STOCK_GO_FORWARD, '_Download', None, 'Download File', self.on_download_btn_clicked)])
 		self.download_actiongrp.set_sensitive(False)
 		self.uimanager.insert_action_group(self.download_actiongrp, 2)
 		
 		self.uimanager.add_ui_from_file('guppy-gtk.xml')
-		
-	
+
 	
 	def customWidgetHandler(self, glade, func_name, widget_name, str1, str2, int1, int2, *args):
-#		print 'glade = ', glade
-#		print 'func_name = ', func_name
-#		print 'widget_name = ', widget_name
-#		print 'str1 = ', str1
-#		print 'str2 = ', str2
-#		print 'int1 = ', int1
-#		print 'int2 = ', int2
-#		print 'args = ',  args
-		
 		handler = getattr(self, func_name)
 		return handler(str1, str2, int1, int2)
 		
@@ -439,6 +491,7 @@ class GuppyWindow:
 			treeview.append_column(col)
 			liststore.set_sort_func(FileSystemModel.SIZE_COL, sort_func, FileSystemModel.SIZE_COL)
 
+		
 	def createMenuBar(self, str1, str2, int1, int2, *args):
 		return self.uimanager.get_widget('/MenuBar')
 		
@@ -456,16 +509,19 @@ class GuppyWindow:
 	
 	def run(self):
 		self.createFileTrees()
+		gtk.gdk.threads_enter()
 		gtk.main()
+		gtk.gdk.threads_leave()
 
 
 	def on_about(self, widget, data=None):	
 		dialog = gtk.AboutDialog()
-		dialog.set_name('Guppy')
-		dialog.set_authors(['Tony Tsui tsui.tony@gmail.com'])
-		dialog.set_copyright('Copyright 2005 Tony Tsui')
-		dialog.set_version('0.0.1')
-		dialog.set_license('GNU Public License')
+		dialog.set_name(APP_NAME)
+		dialog.set_authors(AUTHORS)
+		dialog.set_copyright(COPYRIGHT)
+		dialog.set_version(VERSION)
+		dialog.set_license(LICENSE)
+		dialog.set_website(WEBSITE)
 		dialog.show()
 	
 	def on_column_clicked(self, col, data):
@@ -492,19 +548,20 @@ class GuppyWindow:
 		
 	def on_quit(self, widget, data=None):
 		gtk.main_quit()
+		self.puppy.cancelTransfer()
 		
+	def on_show_file_transfer_toggled(self, widget, data=None):
+		transfer_frame = self.glade_xml.get_widget('transfer_frame')
+		if not transfer_frame.get_property('visible'):
+			self.show_transfer_frame()
+		else:
+			transfer_frame.hide()
+					
 	def on_show_hidden_toggled(self, widget, data=None):
 		self.show_hidden = not self.show_hidden
 		self.pc_liststore.get_model().refilter()
 		self.pvr_liststore.get_model().refilter()
 
-	def on_transfer_dialog_cancel_btn_clicked(self, widget, data=None):
-		self.transferDialogClose()
-		
-	def on_transfer_dialog_delete_event(self, widget, data=None):
-		self.transferDialogClose()
-		return True
-
 	def on_treeview_changed(self, widget, fs_model):
 		model, files = widget.get_selected_rows()
 		
@@ -514,31 +571,29 @@ class GuppyWindow:
 		for path in files:
 			iter = model.get_iter(path)
 			type = model.get_value(iter, FileSystemModel.TYPE_COL)
-			size = model.get_value(iter, FileSystemModel.SIZE_COL)
 			
+			size = model.get_value(iter, FileSystemModel.SIZE_COL)
 			if type != 'd':
 				total_size += convertToBytes(size)
 
 		if total_size > 0:
 			msg = _('Selection Size') + ': ' + humanReadableSize(total_size)
 		else:
-			msg = None
+			msg = _('Selection Size') + ':'
 		
 		if isinstance(fs_model, PCFileSystemModel):
-			if msg:
-				self.pc_total_size_label.set_text(msg)
-				self.upload_actiongrp.set_sensitive(True)
-			else:
-				self.pc_total_size_label.set_text('')
-				self.upload_actiongrp.set_sensitive(False)
+			label = self.pc_total_size_label
+			actiongrp = self.upload_actiongrp
 		else:
-			if msg:
-				self.pvr_total_size_label.set_text(msg)
-				self.download_actiongrp.set_sensitive(True)
-			else:
-				self.pvr_total_size_label.set_text('')
-				self.download_actiongrp.set_sensitive(False)
-		
+			label = self.pvr_total_size_label
+			actiongrp = self.download_actiongrp
+			
+		label.set_text(msg)
+		if total_size > 0:
+			actiongrp.set_sensitive(True)
+		else:
+			actiongrp.set_sensitive(False)
+	
 	def on_treeview_row_activated(self, widget, path, col, fs_model):
 		model = widget.get_model()
 		iter = model.get_iter(path)
@@ -559,19 +614,43 @@ class GuppyWindow:
 	def on_upload_btn_clicked(self, widget, data=None):
 		self.transferFile('upload')
 
-	def transferDialogClose(self):
-		self.puppy.cancelTransfer()
-		self.transfer_dialog.hide()
+	def on_transfer_clear_btn_clicked(self, widget, data=None):
+		# Loop until we get a Queue.Empty exception
+		while True:
+			try:
+				file_transfer = self.transfer_complete_queue.get_nowait()
+			except Queue.Empty:
+				break
 
-		# Update FileSystemModel view				
-		models = [ (self.pc_model, self.pc_treeview), (self.pvr_model, self.pvr_treeview) ]
-		for model, treeview in models:
-			handler_id = treeview.get_data('changed_handler_id')
-			selection = treeview.get_selection()
-			selection.handler_block(handler_id)
-			model.changeDir()
-			selection.handler_unblock(handler_id)
+			progress_box = file_transfer.xml.get_widget('progress_box')
+			# progress_box may be None if the Remove button was clicked on it
+			if progress_box:
+				progress_box.destroy()
 	
+	def on_transfer_close_btn_clicked(self, widget, data=None):
+		self.show_file_transfer_action.set_active(False)
+		self.on_show_file_transfer_toggled(widget)
+
+	def on_transfer_remove_btn_clicked(self, widget, file_transfer):
+		file_transfer.cancel()
+		
+		progress_box = file_transfer.xml.get_widget('progress_box')
+		progress_box.destroy()
+
+	def on_transfer_stop_btn_clicked(self, widget, data=None):
+		self.puppy.cancelTransfer()
+
+	
+	def show_transfer_frame(self):
+		# Set position of pane separator
+		# Give progress frame height of 180 pixels. This should be enough room
+		# to show two file transfers.
+		vpane = self.glade_xml.get_widget('vpane')
+		vpane.set_position(vpane.get_allocation().height - 185)
+		
+		transfer_frame = self.glade_xml.get_widget('transfer_frame')
+		transfer_frame.show()
+		
 	def transferFile(self, direction):
 		if direction == 'download':
 			model, files = self.pvr_treeview.get_selection().get_selected_rows()
@@ -604,23 +683,9 @@ class GuppyWindow:
 			dialog.destroy()
 			if response == gtk.RESPONSE_NO or response == gtk.RESPONSE_DELETE_EVENT:
 				return
-		
-		# Stop free space update timer
-		gobject.source_remove(self.free_space_timeout_id)
-		
-		progress_bar = self.glade_xml.get_widget('transfer_dialog_progressbar')
-		file_label = self.glade_xml.get_widget('transfer_dialog_file_label')
-		file_no_label = self.glade_xml.get_widget('transfer_dialog_file_no_label')
-		from_label = self.glade_xml.get_widget('transfer_dialog_from_label')
-		to_label = self.glade_xml.get_widget('transfer_dialog_to_label')
 
-		dir_label = self.glade_xml.get_widget('transfer_dialog_direction_label1')
-		dir_label.set_markup('<b>' + direction_text + ' ' + _('File') + ':</b>')
-
-		dir_label = self.glade_xml.get_widget('transfer_dialog_direction_label2')
-		dir_label.set_markup('<b>' + direction_text + ':</b>')
-
-		self.transfer_dialog.show()
+		queue_box = self.glade_xml.get_widget('queue_vbox')
+		self.show_transfer_frame()
 
 		file_count = len(files)
 		file_no = 1
@@ -633,14 +698,15 @@ class GuppyWindow:
 				dst_dir = self.pc_model.getCWD()
 				src_file = src_dir + '\\' + file
 				dst_file = dst_dir + '/' + file
+				existing_file = os.access(dst_file, os.F_OK)
 			else:
 				src_dir = self.pc_model.getCWD()
 				dst_dir = self.pvr_model.getCWD()
 				src_file = src_dir + '/' + file
 				dst_file = dst_dir + '\\' + file
+				existing_file = self.pvr_model.find(file)
 
-			if os.access(dst_file, os.F_OK):
-				self.transfer_dialog.hide()
+			if existing_file:
 				msg = _('The file') + ' "' + dst_file + '" ' + _('already exists. Would you like to replace it?')
 				msg2 = _('If you replace an existing file, its contents will be overwritten.')
 				dialog = gtk.MessageDialog(type=gtk.MESSAGE_QUESTION,
@@ -649,48 +715,219 @@ class GuppyWindow:
 				response = dialog.run()
 				dialog.destroy()
 
-				self.transfer_dialog.show()
-
 				if response == gtk.RESPONSE_NO or response == gtk.RESPONSE_DELETE_EVENT:
 					file_no += 1
 					continue
-	
-			if direction == 'download':
-				self.puppy.getFile(src_file, dst_file)
-			else:
-				self.puppy.putFile(src_file, dst_file)
 
+			xml = gtk.glade.XML('guppy.glade', 'progress_box')
+			xml.signal_autoconnect(self)		
+
+			progress_box = xml.get_widget('progress_box')
+			
+			file_label = xml.get_widget('transfer_file_label')
+			from_label = xml.get_widget('transfer_from_label')
+			to_label = xml.get_widget('transfer_to_label')
+	
+			dir_label = xml.get_widget('transfer_direction_label')
+			dir_label.set_markup('<b>' + direction_text + ':</b>')
+	
 			file_label.set_text(file)
 			from_label.set_text(src_dir)
 			to_label.set_text(dst_dir)
-			file_no_label.set_markup('<b>' + str(file_no) + ' ' + _('of') + ' ' + str(file_count) + '</b>')
 
-			file_no += 1
+			queue_box.pack_start(progress_box, expand=False)
+	
+			file_transfer = FileTransfer(direction, src_file, dst_file, xml)
 
-			percent, speed, time = self.puppy.getProgress()
-			while percent != None:
-				progress_bar.set_fraction(float(percent)/100)
-				progress_bar.set_text('(' + time['remaining'] + ' ' + _('Remaining') + ')')
-				while gtk.events_pending():
-					gtk.main_iteration()
-				percent, speed, time = self.puppy.getProgress()
-		
-		self.transferDialogClose()
+			# Connect transfer instance remove button signal handler
+			remove_btn = xml.get_widget('transfer_remove_button')
+			remove_btn.connect('clicked', self.on_transfer_remove_btn_clicked, file_transfer)
 
-		# Restart free space update timer
-		self.free_space_timeout_id = gobject.timeout_add(5000, self.update_free_space)
-		self.update_free_space()
-		
-	def update_free_space(self):		
+			self.transfer_queue.put(file_transfer, True, None)
+
+
+	def update_treeviews(self):		
+		# Update FileSystemModel view				
+		models = [ (self.pc_model, self.pc_treeview), (self.pvr_model, self.pvr_treeview) ]
+		for model, treeview in models:
+			handler_id = treeview.get_data('changed_handler_id')
+			selection = treeview.get_selection()
+			selected_rows = selection.get_selected_rows()
+			selection.handler_block(handler_id)
+			model.changeDir()
+			selection.handler_unblock(handler_id)
+			for path in selected_rows[1]:
+				selection.select_path(path)
+			
+	def update_free_space(self):
 		self.pvr_free_space_label.set_text(_('Free Space') + ': ' + self.pvr_model.freeSpace())
 		self.pc_free_space_label.set_text(_('Free Space') + ': ' + self.pc_model.freeSpace())
 		return True
+
+
 		
+class FileTransfer:
+	def __init__(self, direction, src, dst, xml):
+		self.direction = direction
+		self.src = src
+		self.dst = dst
+		self.xml = xml
+		
+		self.alive = True
+	
+	def isAlive(self):
+		return self.alive
+		
+	def cancel(self):
+		self.alive = False
+	
+	def complete(self):
+		self.alive = False
+		
+class TransferThread(threading.Thread):
+	NUM_OF_RESET_ATTEMPTS = 6
+	NUM_OF_TRANSFER_ATTEMPTS = 2
+	
+	def __init__(self, guppy):
+		threading.Thread.__init__(self)
+		self.guppy = guppy
+		self.file_queue = self.guppy.transfer_queue
+		self.complete_queue = self.guppy.transfer_complete_queue
+
+	def run(self):
+		while True:
+			file_transfer = self.file_queue.get(True, None)
+			if not file_transfer.isAlive():
+				continue
+				
+			direction = file_transfer.direction
+			src_file = file_transfer.src
+			dst_file = file_transfer.dst
+			xml = file_transfer.xml
+			
+			# Make all widgets in progress box active
+			for widget in ['progress_hbox1', 'progress_hbox2']:
+				widget = xml.get_widget(widget)
+				gtk.gdk.threads_enter()
+				widget.set_sensitive(True)
+				gtk.gdk.threads_leave()
+		
+			remove_btn = xml.get_widget('transfer_remove_button')	
+			gtk.gdk.threads_enter()
+			remove_btn.hide()
+			gtk.gdk.threads_leave()
+			
+			stop_btn = xml.get_widget('transfer_stop_button')	
+			gtk.gdk.threads_enter()
+			stop_btn.show()
+			gtk.gdk.threads_leave()
+
+			# Set width of Stop button to be the same as Remove button
+			gtk.gdk.threads_enter()
+			allocation = remove_btn.get_allocation()
+			gtk.gdk.threads_leave()
+			gtk.gdk.threads_enter()
+			stop_btn.set_size_request(allocation.width, allocation.height)
+			gtk.gdk.threads_leave()
+
+			progress_bar = xml.get_widget('transfer_progressbar')
+			
+			if direction == 'download':
+				self.guppy.puppy.getFile(src_file, dst_file)
+			else:
+				self.guppy.puppy.putFile(src_file, dst_file)
+
+			transfer_successful = True
+			transfer_attempt = 1
+			percent = True			
+			while percent != None:
+				try:
+					percent, speed, time = self.guppy.puppy.getProgress()
+				except puppy.PuppyError:
+					# Quit trying to transfer file after a certain number of attempts
+					if transfer_attempt > TransferThread.NUM_OF_TRANSFER_ATTEMPTS:
+						transfer_successful = False
+						break
+					
+					transfer_attempt += 1
+
+					# Try to reset PVR before attempt transfer again
+					reset_successful = False
+					attempts = 0
+					while not reset_successful and attempts < TransferThread.NUM_OF_RESET_ATTEMPTS:
+						reset_successful = self.guppy.puppy.reset()
+						attempts += 1
+
+					if direction == 'download':
+						self.guppy.puppy.getFile(src_file, dst_file)
+					else:
+						self.guppy.puppy.putFile(src_file, dst_file)
+
+					continue
+
+				# Transfer may have completed
+				if percent is None:
+					break
+					
+				gtk.gdk.threads_enter()
+				progress_bar.set_fraction(float(percent)/100)
+				gtk.gdk.threads_leave()
+				
+				gtk.gdk.threads_enter()
+				progress_bar.set_text('(' + time['remaining'] + ' ' + _('Remaining') + ')')
+				gtk.gdk.threads_leave()
+				
+				gtk.gdk.threads_enter()
+				while gtk.events_pending():
+					gtk.main_iteration()
+				gtk.gdk.threads_leave()
+			
+			if transfer_successful:
+				gtk.gdk.threads_enter()
+				progress_bar.set_fraction(1)
+				gtk.gdk.threads_leave()
+				
+				gtk.gdk.threads_enter()
+				progress_bar.set_text(_('Finished'))
+				gtk.gdk.threads_leave()
+			else:
+				gtk.gdk.threads_enter()
+				progress_bar.set_text(_('Transfer Failed'))
+				gtk.gdk.threads_leave()
+				
+
+			# Insensitise all widgets			
+			for widget in ['progress_hbox1', 'progress_hbox2']:
+				widget = xml.get_widget(widget)
+				gtk.gdk.threads_enter()
+				widget.set_sensitive(False)
+				gtk.gdk.threads_leave()
+
+			# Swap Stop button for Remove button
+			gtk.gdk.threads_enter()
+			stop_btn.hide()
+			gtk.gdk.threads_leave()
+			
+			gtk.gdk.threads_enter()
+			remove_btn.show()
+			gtk.gdk.threads_leave()
+
+			file_transfer.complete()
+			
+			# Update file treeviews
+			gtk.gdk.threads_enter()
+			self.guppy.update_treeviews()
+			gtk.gdk.threads_leave()
+			
+			# Put on queue for completed transfers
+			self.complete_queue.put(file_transfer)
+
 if __name__ == "__main__":
 	locale.setlocale(locale.LC_ALL, '')
 	gettext.bindtextdomain(APP_NAME, 'i18n')
 	gettext.textdomain(APP_NAME)
 	gettext.install(APP_NAME, 'i18n', unicode=1)
 	
+	gtk.gdk.threads_init()
 	guppy = GuppyWindow()
 	guppy.run()
