@@ -18,11 +18,19 @@
 import os
 import popen2
 import signal
+import time
+import fcntl
 
 # Set to True for debug output
 DEBUG = False
 
 class Puppy:
+	LOCK_FILE = '/tmp/puppy'
+	
+	# puppy error code for lock failure
+	E_GLOBAL_LOCK = 8
+	E_HDD_NOT_READY = 185
+	
 	def __init__(self):
 		self.cmd = 'puppy'
 		self.turbo = False
@@ -33,7 +41,7 @@ class Puppy:
 		
 		return
 
-	def getDiskSpace(self):		
+	def getDiskSpace(self):
 		args = '-c size'
 
 		output_file = self._execute(args)
@@ -68,13 +76,13 @@ class Puppy:
 	def listDir(self, path=None):
 		args = '-c dir'
 		if path != None:
-			args += ' ' + path
+			args += " '" + path + "'"
 			
 		output_file = self._execute(args)
 
 		output = output_file.readlines()
 		output_file.close()
-		
+
 		listing = []
 		# Parse output of output_file and return it as a list
 		for line in output:
@@ -84,8 +92,11 @@ class Puppy:
 			         "%s %s %s %s" % (entry[2], entry[3], entry[4], entry[6]),
 			         entry[1] ]
 			listing.append(item)
-			
-		if self.getStatus() != 0:
+		
+		status = self.getStatus()	
+		if status == Puppy.E_GLOBAL_LOCK:
+			raise PuppyBusyError(str(output))
+		elif status != 0:
 			raise PuppyError("listDir failed. puppy returned: " + str(output))
 		
 		return listing
@@ -94,23 +105,31 @@ class Puppy:
 	def getFile(self, src_file, dest_file=None):
 		args = '-c get' + ' "' + src_file + '"'
 		if dest_file != None:
-			args += ' "' + dest_file + '"'
+			args += " '" + dest_file + "'"
 		else:
-			args += ' "' + os.path.basename(src_file) + '"'
+			args += " '" + os.path.basename(src_file) + "'"
 			
 		self.progress_output = self._execute(args)
-		
+
+		status = self.getStatus(wait=False)
+		if status != 0 and status != -1:
+			raise PuppyError("getFile failed. puppy returned: " + str(self.progress_output))
+
 		return
 
 	def putFile(self, src_file, dest_file=None):
 		args = '-c put' + ' "' + src_file + '"'
 		if dest_file != None:
-			args += ' "' + dest_file + '"'
+			args += " '" + dest_file + "'"
 		else:
-			args += ' "' + os.path.basename(src_file) + '"'
+			args += " '" + os.path.basename(src_file) + "'"
 			
 		self.progress_output = self._execute(args)
 		
+		status = self.getStatus(wait=False)
+		if status != 0 and status != -1:
+			raise PuppyError("putFile failed. puppy returned: " + str(self.progress_output))
+			
 		return
 
 	def makeDir(self, dirname):
@@ -161,6 +180,8 @@ class Puppy:
 	def getProgress(self):
 		exit_status = self.getStatus(wait=False)
 		if exit_status != -1:
+			# Raise exception if puppy did not exit successfully or because of
+			# SIGTERM.
 			if exit_status != 0 and exit_status != 15:
 				raise PuppyError("Transfer failed")
 			
@@ -200,6 +221,32 @@ class Puppy:
 		
 	def setTurbo(self, value):
 		self.turbo = value
+	
+	def reset(self):	
+		args = '-c cancel'
+			
+		output_file = self._execute(args)
+
+		output = output_file.readlines()
+		output_file.close()
+		
+		if self.getStatus() != 0:
+			return False
+		
+		return True
+
+	def _anotherPuppyActive(self):
+		lock_file = open(Puppy.LOCK_FILE, 'a')
+		
+		try:
+			fcntl.flock(lock_file, fcntl.LOCK_EX | fcntl.LOCK_NB)
+			result = False
+		except:
+			result = True
+			
+		lock_file.close()
+	
+		return result
 		
 	def _execute(self, args):
 		cmd = self.cmd
@@ -209,10 +256,21 @@ class Puppy:
 		
 		if DEBUG:
 			print 'cmd = ', cmd
-		
+
+		if self._anotherPuppyActive():
+			raise PuppyBusyError('Can not get exclusive lock on ' + Puppy.LOCK_FILE)
+			
 		self.popen_obj = popen2.Popen4(cmd)
 		self.popen_obj.tochild.close()
 		
+		status = self.getStatus(wait=False)
+		if status == Puppy.E_GLOBAL_LOCK:
+			raise PuppyBusyError(self.popen_obj.fromchild.readlines())
+		elif status == Puppy.E_HDD_NOT_READY:
+			time.sleep(1)
+			self.popen_obj = popen2.Popen4(cmd)
+			self.popen_obj.tochild.close()
+			
 		return self.popen_obj.fromchild
 
 class PuppyError(Exception):
@@ -220,3 +278,10 @@ class PuppyError(Exception):
 		self.value = value
 	def __str__(self):
 		return repr(self.value)
+
+class PuppyBusyError(Exception):
+	def __init__(self, value):
+		self.value = value
+	def __str__(self):
+		return repr(self.value)
+	
